from __future__ import division

import sys
from math import trunc, exp, log, ceil, floor
import matplotlib.pyplot as plt
import matplotlib.axis as axis
import matplotlib.ticker as mtick
from matplotlib import rc

import random

from schedcat.model.LLC_params import *
import schedcat.model.ohc as ohc
import schedcat.model.mcx_tasks as ts
import numpy as np
from matplotlib import cm

from schedcat.model.LLCLayout import *

def uniform_int(minval, maxval):
    "Create a function that draws ints uniformly from {minval, ..., maxval}"
    def _draw():
        return random.randint(minval, maxval)
    return _draw

def uniform(minval, maxval):
    "Create a function that draws floats uniformly from [minval, maxval]"
    def _draw():
        return random.uniform(minval, maxval)
    return _draw

def log_uniform(minval, maxval):
    "Create a function that draws floats log-uniformly from [minval, maxval]"
    def _draw():
        return exp(random.uniform(log(minval), log(maxval))) 
    return _draw

def log_uniform_int(minval, maxval):
    "Create a function that draws ints log-uniformly from {minval, ..., maxval}"
    draw_float = log_uniform(minval, maxval + 1)
    def _draw():
        val = int(draw_float())
        val = max(minval, val)
        val = min(maxval, val)
        return val
    return _draw

def uniform_choice(choices):
    "Create a function that draws uniformly elements from choices"
    selector = uniform_int(0, len(choices) - 1)
    def _draw():
        return choices[selector()]
    return _draw

def geometric(p, include_zero = False):
    if include_zero:
        rounder = floor
    else:
        rounder = ceil
    def _draw():
        return rounder(log(1-random.random()) / log(1-p))
    return _draw

def truncate(minval, maxval):
    def _limit(fun):
        def _f(*args, **kargs):
            val = fun(*args, **kargs)
            return min(maxval, max(minval, val))
        return _f
    return _limit

def redraw(minval, maxval):
    def _redraw(dist):
        def _f(*args, **kargs):
            in_range = False
            while not in_range:
                val = dist(*args, **kargs)
                in_range = minval <= val <= maxval
            return val
        return _f
    return _redraw

def exponential(minval, maxval, mean, limiter=redraw):
    """Create a function that draws floats from an exponential
    distribution with expected value 'mean'. If a drawn value is less
    than minval or greater than maxval, then either another value is
    drawn (if limiter=redraw) or the drawn value is set to minval or
    maxval (if limiter=truncate)."""
    def _draw():
        return random.expovariate(1.0 / mean)
    return limiter(minval, maxval)(_draw)

def multimodal(weighted_distributions):
    """Create a function that draws values from several distributions
    with probability according to the given weights in a list of
    (distribution, weight) pairs."""
    total_weight = sum([w for (d, w) in weighted_distributions])
    selector = uniform(0, total_weight)
    def _draw():
        x = selector()
        wsum = 0
        for (d, w) in weighted_distributions:
            wsum += w
            if wsum >= x:
                return d()
        assert False # should never drop off
    return _draw

def restrict(bounds, wMin, wMax):
    interval = bounds[1]-bounds[0]
    return [bounds[0]+wMin*interval, bounds[0]+wMax*interval]

def plot(cost, period, atlevel, file):
    print "plotting"
    fig = plt.figure()
    
    X = np.arange(0, 17, 1)
    Y = np.arange(0, 17, 1)
    X, Y = np.meshgrid(X, Y)
    Z = []
    
    for s in range(17):
        Z.append([])
        for w in range(17):
            Z[s].append(cost[atlevel][s][w])
    
    print X
    print Y
    print Z
    
    ax = fig.add_subplot(1, 1, 1, projection='3d')
    ax.plot_surface(X, Y, Z, rstride=1, cstride=1)
    
    ax.set_zlim(0, 7000)
    ax.view_init(40, 45)
    plt.savefig("../PETS/" + str(file) + ".png", bbox_inches='tight', pad_inches=0.1)

    plt.close()
    
    
    print "done"


class Custom:
    pass
    
class TaskGeneratorMCX(object):
    """Sporadic task generator"""

    #def __init__(self, period, util, util_dist, mem_util, avg_percent, a_inflation, pfuncs, pet_sens, maxWays = 16, maxSets = 16, deadline=lambda x, y: y,):
    def __init__(self, period, util, util_dist, mem_util, avg_percent, a_inflation, pfuncs, maxWays = 16, maxSets = 16, deadline=lambda x, y: y,):
        """Creates TaskGenerator based on a given a period and
        utilization distributions."""
        self.period    = period
        self.util      = util
        self.deadline  = deadline
        self.mW = maxWays
        self.mS = maxSets
        self.util_dist = util_dist
        self.mem_util = mem_util
        self.P = pfuncs
        self.avg_percent = avg_percent
        self.a_inflation = a_inflation
        
        #self.util_var = util_var
        #self.pet_sens = pet_sens

	#self.io_buff_size = io_buff_size
	#self.io_task_dist = io_task_dist
	#self.mem_footprint = mem_footprint

	#self.net_exec = net_exec
	#self.net_period = net_period
	#self.net_dist = net_dist
	#self.has_cam = has_cam
    
#     def Pv(self, S, W, Psam):
#         return (1-Psam.kmin)*exp(-S*Psam.ks-W*Psam.kw-S*W*Psam.kw)+Psam.kmin

    def Pv(self, sz, Psam):
        return (1-Psam.kmin)*exp(-sz*Psam.ks)+Psam.kmin
        
        
    def Pv(self, sz, Psam):
        return (1-Psam.kmin)*exp(-sz*Psam.ks)+Psam.kmin
    
    def Psample(self, P):
        Psamp = Custom()
        
        Psamp.ks = P.ks()
        #Psamp.kw = P.kw()
        #Psamp.kp = P.kp()
        Psamp.kmin = P.kmin()
        Psamp.infMax = P.infMax()
        Psamp.aMax = P.aMax()
        Psamp.decay = P.decay()
	Psamp.way_ind = P.way_ind()
	Psamp.c_scale = P.c_scale()
        Psamp.c_shared = P.c_shared()
        return Psamp

    def tasks(self, tot_mc2_util, max_tasks=None, max_util=None, squeeze=False,
              time_conversion=trunc, a_inflate = 1.5, level = 'C'):
        """Generate a sequence of tasks until either max_tasks is reached
        or max_util is reached. If max_util would be exceeded and squeeze is
        true, then the last-generated task's utilization is scaled to exactly
        match max_util. Otherwise, the last-generated task is discarded.
        time_conversion is used to convert the generated (non-integral) values
        into integral task parameters.
        """
        count = 0
        usum  = 0
	first = True
        while ((max_tasks is None or count < max_tasks) and
               (max_util is None  or usum  < max_util)):
	    period = 0
	    #if first and self.has_cam and level == 'B': #Camera is a level B task with fixed period
	    #	period	 = 96*1000
	    #else:
            period   = self.period[level]()
            util     = self.util[level]()
            cost     = period * util
            deadline = self.deadline(cost, period)
            # scale as required
            period   = max(1, int(time_conversion(period)))
            cost     = max(1, int(time_conversion(cost)))
            deadline = max(1, int(time_conversion(deadline)))
	    #footprint = self.mem_footprint[level]()
            #tsk_pet_sens = {'way' : self.pet_sens['way'](), 'color' : self.pet_sens['color']()}
            util = cost / period
            count  += 1
            usum   += util
            if max_util and usum > max_util:
                if squeeze:
                    # make last task fit exactly
                    util -= (usum - max_util)
                    cost = trunc(period * util)
                else:
                    break
            
            if cost > 0:
                wss = self.mem_util()*cost/(1.5*ohc.CPMD) #of 1-way X 1-color LLC area "units" 
                    
                # ----Generate A and B Partition Functions---- #
                PA = self.Psample(self.P[level]['A'])
                               
                #scales = {'A': [[self.Pv(min(s*w, wss), PA) for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]}
                
                # First set up A
                #for i in range(0, self.mW+1):
                #    scales['A'][0][i] = 1.0
                #for i in range(0, self.mS+1):
                #    scales['A'][i][0] = 1.0
                        
                #minA = scales['A'][self.mS][self.mW]
                #for s in range(0, self.mS+1):
                #    for w in range(0, self.mW+1):
                #        scales['A'][s][w] /= minA
            
                # Then B
                #scales['B'] = [[scales['A'][s][w]/a_inflate for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
                
                #costs = dict()
                
                #for key in scales.keys():
                #    costs[key] = [[scale*cost for scale in scales_s] for scales_s in scales[key]]
                    
                
                #backload = tot_mc2_util - costs['A'][self.mW][self.mS]/period
                
                #c_percent = self.avg_percent(backload, level)
                
                #costs['Ciso'] = [[costs['B'][s][w]*c_percent for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
                
                #highCcost = costs['Ciso'][0][0]
                #lowCcost = costs['Ciso'][self.mS][self.mW]
                
                #lowCshared = lowCcost+random.uniform(0.3, 0.7)*(highCcost-lowCcost)
                
                #costs['Cshared'] = [[highCcost-(float(s*w)/float(self.mW*self.mS))*(highCcost-lowCshared) for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
                
                #if not level is 'C':
                #    costs['C'] = [[costs['Ciso'][s][w] for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
                
                tsk = ts.SporadicTaskMCX(base_cost = cost, period = period, wss = wss, level = level, deadline = deadline, sample = PA)
                
		tsk.device = 'none'
                yield tsk
            else:
                break

    def pick_B(self):
        return random.random() < 0.2
        
    def make_task_set(self, max_util):
    	a_inf = self.a_inflation()
	utilA_A = self.util_dist['A']()
	utilB_A = self.util_dist['B']()
	utilC_A = self.util_dist['C']()
	normFactor = max_util/(utilA_A + utilB_A + utilC_A)
	utilA_A *= normFactor
	utilB_A *= normFactor
	utilC_A *= normFactor

	alltasks = []
	alltasks.extend(self.tasks(squeeze = True, tot_mc2_util = max_util, level = 'A', max_util = utilA_A, a_inflate = a_inf))
	lvl_A_len = len(alltasks)
	alltasks.extend(self.tasks(squeeze = True, tot_mc2_util = max_util, level = 'B', max_util = utilB_A, a_inflate = a_inf))
	lvl_B_len = len(alltasks) - lvl_A_len
	alltasks.extend(self.tasks(squeeze = True, tot_mc2_util = max_util, level = 'C', max_util = utilC_A, a_inflate = a_inf))
	lvl_C_len = len(alltasks) - lvl_A_len - lvl_B_len

	ts_mcx = ts.TaskSystemMCX(alltasks)
	ts_mcx.a_inflate_choice = a_inf

        return ts_mcx  
